# python lib
from subprocess import PIPE
from mock import patch, call, Mock

# local lib
from jupyterlab_git.git import Git


@patch('subprocess.Popen')
def test_detailed_log(mock_subproc_popen):
    # Given
    process_output = [
        'f29660a (HEAD, origin/feature) Commit message',
        '10      3       notebook_without_spaces.ipynb',
        '11      4       Notebook with spaces.ipynb',
        '12      5       path/notebook_without_spaces.ipynb',
        '13      6       path/Notebook with spaces.ipynb',
        ' notebook_without_spaces.ipynb      | 13 ++++++++---',
        ' Notebook with spaces.ipynb         | 15 +++++++++----',
        ' path/notebook_without_spaces.ipynb | 17 ++++++++++-----',
        ' path/Notebook with spaces.ipynb    | 19 +++++++++++------',
        ' 4 files changed, 46 insertions(+), 18 deletions(-)'
    ]
    process_mock = Mock(returncode=0)
    process_mock.communicate.side_effect = [
        ('\n'.join(process_output).encode('utf-8'), ''.encode('utf-8')),
    ]
    mock_subproc_popen.return_value = process_mock

    expected_response = {
        'code': 0,
        'modified_file_note': ' 4 files changed, 46 insertions(+), 18 deletions(-)',
        'modified_files_count': '4',
        'number_of_insertions': '46',
        'number_of_deletions': '18',
        'modified_files': [
            {
                'modified_file_path': 'notebook_without_spaces.ipynb',
                'modified_file_name': 'notebook_without_spaces.ipynb',
                'insertion': '10',
                'deletion': '3'
            },
            {
                'modified_file_path': 'Notebook with spaces.ipynb',
                'modified_file_name': 'Notebook with spaces.ipynb',
                'insertion': '11',
                'deletion': '4'
            },
            {
                'modified_file_path': 'path/notebook_without_spaces.ipynb',
                'modified_file_name': 'notebook_without_spaces.ipynb',
                'insertion': '12',
                'deletion': '5'
            },
            {
                'modified_file_path': 'path/Notebook with spaces.ipynb',
                'modified_file_name': 'Notebook with spaces.ipynb',
                'insertion': '13',
                'deletion': '6'
            }
        ]
    }

    # When
    actual_response = Git(root_dir='/bin').detailed_log(
        selected_hash='f29660a2472e24164906af8653babeb48e4bf2ab',
        current_path='test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'log', '-1', '--stat', '--numstat', '--oneline', 'f29660a2472e24164906af8653babeb48e4bf2ab'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate(),
    ], any_order=False)

    assert expected_response == actual_response



from subprocess import PIPE, STDOUT, CalledProcessError

from mock import Mock, call, patch
import pytest
from tornado.web import HTTPError


from jupyterlab_git.git import Git


def test_changed_files_invalid_input():
    with pytest.raises(HTTPError):
        actual_response = Git(root_dir="/bin").changed_files(
            base="64950a634cd11d1a01ddfedaeffed67b531cb11e"
        )


@patch("subprocess.check_output")
def test_changed_files_single_commit(mock_call):
    # Given
    mock_call.return_value = b"file1.ipynb\nfile2.py"

    # When
    actual_response = Git(root_dir="/bin").changed_files(
        single_commit="64950a634cd11d1a01ddfedaeffed67b531cb11e"
    )

    # Then
    mock_call.assert_called_with(
        ["git", "diff", "64950a634cd11d1a01ddfedaeffed67b531cb11e^!", "--name-only"],
        cwd="/bin",
        stderr=STDOUT,
    )
    assert {"code": 0, "files": ["file1.ipynb", "file2.py"]} == actual_response


@patch("subprocess.check_output")
def test_changed_files_working_tree(mock_call):
    # Given
    mock_call.return_value = b"file1.ipynb\nfile2.py"

    # When
    actual_response = Git(root_dir="/bin").changed_files(base="WORKING", remote="HEAD")

    # Then
    mock_call.assert_called_with(
        ["git", "diff", "HEAD", "--name-only"], cwd="/bin", stderr=STDOUT
    )
    assert {"code": 0, "files": ["file1.ipynb", "file2.py"]} == actual_response


@patch("subprocess.check_output")
def test_changed_files_index(mock_call):
    # Given
    mock_call.return_value = b"file1.ipynb\nfile2.py"

    # When
    actual_response = Git(root_dir="/bin").changed_files(base="INDEX", remote="HEAD")

    # Then
    mock_call.assert_called_with(
        ["git", "diff", "--staged", "HEAD", "--name-only"], cwd="/bin", stderr=STDOUT
    )
    assert {"code": 0, "files": ["file1.ipynb", "file2.py"]} == actual_response


@patch("subprocess.check_output")
def test_changed_files_two_commits(mock_call):
    # Given
    mock_call.return_value = b"file1.ipynb\nfile2.py"

    # When
    actual_response = Git(root_dir="/bin").changed_files(
        base="HEAD", remote="origin/HEAD"
    )

    # Then
    mock_call.assert_called_with(
        ["git", "diff", "HEAD", "origin/HEAD", "--name-only"], cwd="/bin", stderr=STDOUT
    )
    assert {"code": 0, "files": ["file1.ipynb", "file2.py"]} == actual_response


@patch("subprocess.check_output")
def test_changed_files_git_diff_error(mock_call):
    # Given
    mock_call.side_effect = CalledProcessError(128, "cmd", b"error message")

    # When
    actual_response = Git(root_dir="/bin").changed_files(
        base="HEAD", remote="origin/HEAD"
    )

    # Then
    mock_call.assert_called_with(
        ["git", "diff", "HEAD", "origin/HEAD", "--name-only"], cwd="/bin", stderr=STDOUT
    )
    assert {"code": 128, "message": "error message"} == actual_response



from subprocess import PIPE

from mock import patch, call, Mock

from jupyterlab_git.git import Git


@patch('subprocess.Popen')
@patch('os.environ', {'TEST': 'test'})
def test_git_pull_fail(mock_subproc_popen):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': ('output', 'Authentication failed'.encode('utf-8')),
        'returncode': 1
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').pull('test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'pull', '--no-commit'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '0'},
        ),
        call().communicate()
    ])
    assert {'code': 1, 'message': 'Authentication failed'} == actual_response

@patch('jupyterlab_git.git.GitAuthInputWrapper')
@patch('os.environ', {'TEST': 'test'})
def test_git_pull_with_auth_fail(mock_GitAuthInputWrapper):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': "remote: Invalid username or password.\r\nfatal: Authentication failed for 'repo_url'".encode('utf-8'),
        'returncode': 1
    }
    process_mock.configure_mock(**attrs)
    mock_GitAuthInputWrapper.return_value = process_mock

    # When
    auth = {
        'username' : 'asdf', 
        'password' : 'qwerty'
    }
    actual_response = Git(root_dir='/bin').pull('test_curr_path', auth)


    # Then
    mock_GitAuthInputWrapper.assert_has_calls([
        call(
            command = 'git pull --no-commit',
            cwd='/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '1'},
            username = 'asdf',
            password = 'qwerty'
        ),
        call().communicate()
    ])
    assert {'code': 1, 'message': "remote: Invalid username or password.\r\nfatal: Authentication failed for 'repo_url'"} == actual_response

@patch('subprocess.Popen')
@patch('os.environ', {'TEST': 'test'})
def test_git_pull_success(mock_subproc_popen):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': ('output', ''.encode('utf-8')),
        'returncode': 0
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').pull('test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'pull', '--no-commit'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '0'},
        ),
        call().communicate()
    ])
    assert {'code': 0} == actual_response

@patch('jupyterlab_git.git.GitAuthInputWrapper')
@patch('os.environ', {'TEST': 'test'})
def test_git_pull_with_auth_success(mock_GitAuthInputWrapper):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': ('output', ''.encode('utf-8')),
        'returncode': 0
    }
    process_mock.configure_mock(**attrs)
    mock_GitAuthInputWrapper.return_value = process_mock

    # When
    auth = {
        'username' : 'asdf', 
        'password' : 'qwerty'
    }
    actual_response = Git(root_dir='/bin').pull('test_curr_path', auth)

    # Then
    mock_GitAuthInputWrapper.assert_has_calls([
        call(
            command = 'git pull --no-commit',
            cwd='/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '1'},
            username = 'asdf',
            password = 'qwerty'
        ),
        call().communicate()
    ])
    assert {'code': 0} == actual_response

@patch('subprocess.Popen')
@patch('os.environ', {'TEST': 'test'})
def test_git_push_fail(mock_subproc_popen):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': ('output', 'Authentication failed'.encode('utf-8')),
        'returncode': 1
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').push('test_origin', 'HEAD:test_master', 'test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'push', 'test_origin', 'HEAD:test_master'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '0'},
        ),
        call().communicate()
    ])
    assert {'code': 1, 'message': 'Authentication failed'} == actual_response

@patch('jupyterlab_git.git.GitAuthInputWrapper')
@patch('os.environ', {'TEST': 'test'})
def test_git_push_with_auth_fail(mock_GitAuthInputWrapper):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': "remote: Invalid username or password.\r\nfatal: Authentication failed for 'repo_url'".encode('utf-8'),
        'returncode': 1
    }
    process_mock.configure_mock(**attrs)
    mock_GitAuthInputWrapper.return_value = process_mock

    # When
    auth = {
        'username' : 'asdf', 
        'password' : 'qwerty'
    }
    actual_response = Git(root_dir='/bin').push('test_origin', 'HEAD:test_master', 'test_curr_path', auth)

    # Then
    mock_GitAuthInputWrapper.assert_has_calls([
        call(
            command = 'git push test_origin HEAD:test_master',
            cwd='/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '1'},
            username = 'asdf',
            password = 'qwerty'
        ),
        call().communicate()
    ])
    assert {'code': 1, 'message': "remote: Invalid username or password.\r\nfatal: Authentication failed for 'repo_url'"} == actual_response


@patch('subprocess.Popen')
@patch('os.environ', {'TEST': 'test'})
def test_git_push_success(mock_subproc_popen):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': ('output', 'does not matter'.encode('utf-8')),
        'returncode': 0
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').push('.', 'HEAD:test_master', 'test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'push', '.', 'HEAD:test_master'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '0'},
        ),
        call().communicate()
    ])
    assert {'code': 0} == actual_response

@patch('jupyterlab_git.git.GitAuthInputWrapper')
@patch('os.environ', {'TEST': 'test'})
def test_git_push_with_auth_success(mock_GitAuthInputWrapper):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': 'does not matter'.encode('utf-8'),
        'returncode': 0
    }
    process_mock.configure_mock(**attrs)
    mock_GitAuthInputWrapper.return_value = process_mock

    # When
    auth = {
        'username' : 'asdf', 
        'password' : 'qwerty'
    }
    actual_response = Git(root_dir='/bin').push('.', 'HEAD:test_master', 'test_curr_path', auth)

    # Then
    mock_GitAuthInputWrapper.assert_has_calls([
        call(
            command = 'git push . HEAD:test_master',
            cwd='/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '1'},
            username = 'asdf',
            password = 'qwerty'
        ),
        call().communicate()
    ])
    assert {'code': 0} == actual_response


"""
Module with all the individual handlers, which execute git commands and return the results to the frontend.
"""
import json

from notebook.utils import url_path_join as ujoin
from notebook.base.handlers import APIHandler


class GitHandler(APIHandler):
    """
    Top-level parent class.
    """

    @property
    def git(self):
        return self.settings["git"]


class GitCloneHandler(GitHandler):
    def post(self):
        """
        Handler for the `git clone`

        Input format:
            {
              'current_path': 'current_file_browser_path',
              'repo_url': 'https://github.com/path/to/myrepo',
              OPTIONAL 'auth': '{ 'username': '<username>',
                                  'password': '<password>'
                                }'
            }
        """
        data = json.loads(self.request.body.decode('utf-8'))
        response = self.git.clone(data['current_path'], data['clone_url'], data.get('auth', None))
        self.finish(json.dumps(response))


class GitAllHistoryHandler(GitHandler):
    """
    Parent handler for all four history/status git commands:
    1. git show_top_level
    2. git branch
    3. git log
    4. git status
    Called on refresh of extension's widget
    """

    def post(self):
        """
        POST request handler, calls individual handlers for
        'git show_top_level', 'git branch', 'git log', and 'git status'
        """
        body = self.get_json_body()
        current_path = body["current_path"]
        history_count = body["history_count"]

        show_top_level = self.git.show_top_level(current_path)
        if show_top_level["code"] != 0:
            self.finish(json.dumps(show_top_level))
        else:
            branch = self.git.branch(current_path)
            log = self.git.log(current_path, history_count)
            status = self.git.status(current_path)

            result = {
                "code": show_top_level["code"],
                "data": {
                    "show_top_level": show_top_level,
                    "branch": branch,
                    "log": log,
                    "status": status,
                },
            }
            self.finish(json.dumps(result))


class GitShowTopLevelHandler(GitHandler):
    """
    Handler for 'git rev-parse --show-toplevel'.
    Displays the git root directory inside a repository.
    """

    def post(self):
        """
        POST request handler, displays the git root directory inside a repository.
        """
        current_path = self.get_json_body()["current_path"]
        result = self.git.show_top_level(current_path)
        self.finish(json.dumps(result))


class GitShowPrefixHandler(GitHandler):
    """
    Handler for 'git rev-parse --show-prefix'.
    Displays the prefix path of a directory in a repository,
    with respect to the root directory.
    """

    def post(self):
        """
        POST request handler, displays the prefix path of a directory in a repository,
        with respect to the root directory.
        """
        current_path = self.get_json_body()["current_path"]
        result = self.git.show_prefix(current_path)
        self.finish(json.dumps(result))


class GitStatusHandler(GitHandler):
    """
    Handler for 'git status --porcelain', fetches the git status.
    """

    def get(self):
        """
        GET request handler, shows file status, used in refresh method.
        """
        self.finish(
            json.dumps(
                {"add_all": "check", "filename": "filename", "top_repo_path": "path"}
            )
        )

    def post(self):
        """
        POST request handler, fetches the git status.
        """
        current_path = self.get_json_body()["current_path"]
        result = self.git.status(current_path)
        self.finish(json.dumps(result))


class GitLogHandler(GitHandler):
    """
    Handler for 'git log --pretty=format:%H-%an-%ar-%s'.
    Fetches Commit SHA, Author Name, Commit Date & Commit Message.
    """

    def post(self):
        """
        POST request handler,
        fetches Commit SHA, Author Name, Commit Date & Commit Message.
        """
        body = self.get_json_body()
        current_path = body["current_path"]
        history_count = body.get("history_count", 25)
        result = self.git.log(current_path, history_count)
        self.finish(json.dumps(result))


class GitDetailedLogHandler(GitHandler):
    """
    Handler for 'git log -1 --stat --numstat --oneline' command.
    Fetches file names of committed files, Number of insertions &
    deletions in that commit.
    """

    def post(self):
        """
        POST request handler, fetches file names of committed files, Number of
        insertions & deletions in that commit.
        """
        data = self.get_json_body()
        selected_hash = data["selected_hash"]
        current_path = data["current_path"]
        result = self.git.detailed_log(selected_hash, current_path)
        self.finish(json.dumps(result))


class GitDiffHandler(GitHandler):
    """
    Handler for 'git diff --numstat'. Fetches changes between commits & working tree.
    """

    def post(self):
        """
        POST request handler, fetches differences between commits & current working
        tree.
        """
        top_repo_path = self.get_json_body()["top_repo_path"]
        my_output = self.git.diff(top_repo_path)
        self.finish(my_output)


class GitBranchHandler(GitHandler):
    """
    Handler for 'git branch -a'. Fetches list of all branches in current repository
    """

    def post(self):
        """
        POST request handler, fetches all branches in current repository.
        """
        current_path = self.get_json_body()["current_path"]
        result = self.git.branch(current_path)
        self.finish(json.dumps(result))


class GitAddHandler(GitHandler):
    """
    Handler for git add <filename>'.
    Adds one or all files to the staging area.
    """

    def get(self):
        """
        GET request handler, adds files to the staging area.
        """
        self.finish(
            json.dumps(
                {"add_all": "check", "filename": "filename", "top_repo_path": "path"}
            )
        )

    def post(self):
        """
        POST request handler, adds one or all files into the staging area.
        """
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        if data["add_all"]:
            my_output = self.git.add_all(top_repo_path)
        else:
            filename = data["filename"]
            my_output = self.git.add(filename, top_repo_path)
        self.finish(my_output)


class GitAddAllUnstagedHandler(GitHandler):
    """
    Handler for 'git add -u'. Adds ONLY all unstaged files, does not touch
    untracked or staged files.
    """

    def post(self):
        """
        POST request handler, adds all the changed files.
        """
        self.finish(
            self.git.add_all_unstaged(
                self.get_json_body()["top_repo_path"]
            )
        )


class GitAddAllUntrackedHandler(GitHandler):
    """
    Handler for 'echo "a\n*\nq\n" | git add -i'. Adds ONLY all
    untracked files, does not touch unstaged or staged files.
    """

    def post(self):
        """
        POST request handler, adds all the untracked files.
        """
        self.finish(
            self.git.add_all_untracked(
                self.get_json_body()["top_repo_path"]
            )
        )


class GitResetHandler(GitHandler):
    """
    Handler for 'git reset <filename>'.
    Moves one or all files from the staged to the unstaged area.
    """

    def post(self):
        """
        POST request handler,
        moves one or all files from the staged to the unstaged area.
        """
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        if data["reset_all"]:
            my_output = self.git.reset_all(top_repo_path)
        else:
            filename = data["filename"]
            my_output = self.git.reset(filename, top_repo_path)
        self.finish(my_output)


class GitDeleteCommitHandler(GitHandler):
    """
    Handler for 'git revert --no-commit <SHA>'.
    Deletes the specified commit from the repository, leaving history intact.
    """

    def post(self):
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        commit_id = data["commit_id"]
        output = self.git.delete_commit(commit_id, top_repo_path)
        self.finish(output)


class GitResetToCommitHandler(GitHandler):
    """
    Handler for 'git reset --hard <SHA>'.
    Deletes all commits from head to the specified commit, making the specified commit the new head.
    """

    def post(self):
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        commit_id = data["commit_id"]
        output = self.git.reset_to_commit(commit_id, top_repo_path)
        self.finish(output)


class GitCheckoutHandler(GitHandler):
    """
    Handler for 'git checkout <branchname>'. Changes the current working branch.
    """

    def post(self):
        """
        POST request handler, changes between branches.
        """
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        if data["checkout_branch"]:
            if data["new_check"]:
                my_output = self.git.checkout_new_branch(
                    data["branchname"], top_repo_path
                )
            else:
                my_output = self.git.checkout_branch(data["branchname"], top_repo_path)
        elif data["checkout_all"]:
            my_output = self.git.checkout_all(top_repo_path)
        else:
            my_output = self.git.checkout(data["filename"], top_repo_path)
        self.finish(my_output)


class GitCommitHandler(GitHandler):
    """
    Handler for 'git commit -m <message>'. Commits files.
    """

    def post(self):
        """
        POST request handler, commits files.
        """
        data = self.get_json_body()
        top_repo_path = data["top_repo_path"]
        commit_msg = data["commit_msg"]
        my_output = self.git.commit(commit_msg, top_repo_path)
        self.finish(my_output)


class GitUpstreamHandler(GitHandler):
    def post(self):
        """
        Handler for the `git rev-parse --abbrev-ref $CURRENT_BRANCH_NAME@{upstream}` on the repo. Used to check if there
        is a upstream branch defined for the current Git repo (and a side-effect is disabling the Git push/pull actions)

        Input format:
            {
              'current_path': 'current_file_browser_path',
            }
        """
        current_path = self.get_json_body()["current_path"]
        current_branch = self.git.get_current_branch(current_path)
        upstream = self.git.get_upstream_branch(current_path, current_branch)
        self.finish(json.dumps({"upstream": upstream}))


class GitPullHandler(GitHandler):
    """
    Handler for 'git pull'. Pulls files from a remote branch.
    """

    def post(self):
        """
        POST request handler, pulls files from a remote branch to your current branch.
        """
        data = self.get_json_body()
        response = self.git.pull(data['current_path'], data.get('auth', None))

        self.finish(json.dumps(response))


class GitPushHandler(GitHandler):
    """
    Handler for 'git push <first-branch> <second-branch>.
    Pushes committed files to a remote branch.
    """

    def post(self):
        """
        POST request handler,
        pushes committed files from your current branch to a remote branch
        """
        data = self.get_json_body()
        current_path = data['current_path']

        current_local_branch = self.git.get_current_branch(current_path)
        current_upstream_branch = self.git.get_upstream_branch(
            current_path, current_local_branch
        )

        if current_upstream_branch and current_upstream_branch.strip():
            upstream = current_upstream_branch.split("/")
            if len(upstream) == 1:
                # If upstream is a local branch
                remote = "."
                branch = ":".join(["HEAD", upstream[0]])
            else:
                # If upstream is a remote branch
                remote = upstream[0]
                branch = ":".join(["HEAD", upstream[1]])

            response = self.git.push(remote, branch, current_path, data.get('auth', None))

        else:
            response = {
                "code": 128,
                "message": "fatal: The current branch {} has no upstream branch.".format(
                    current_local_branch
                ),
            }
        self.finish(json.dumps(response))


class GitInitHandler(GitHandler):
    """
    Handler for 'git init'. Initializes a repository.
    """

    def post(self):
        """
        POST request handler, initializes a repository.
        """
        current_path = self.get_json_body()["current_path"]
        my_output = self.git.init(current_path)
        self.finish(my_output)



class GitChangedFilesHandler(GitHandler):

    def post(self):
        self.finish(
            json.dumps(
                self.git.changed_files(**self.get_json_body())
            )
        )


class GitConfigHandler(GitHandler):
    """
    Handler for 'git config' commands
    """

    def post(self):
        """
        POST get (if no options are passed) or set configuration options
        """
        data = self.get_json_body()
        top_repo_path = data["path"]
        options = data.get("options", {})
        response = self.git.config(top_repo_path, **options)

        if response["code"] != 0:
            self.set_status(500)
        else:
            self.set_status(201)
        self.finish(json.dumps(response))


class GitServerRootHandler(GitHandler):

    def get(self):
        # Similar to https://github.com/jupyter/nbdime/blob/master/nbdime/webapp/nb_server_extension.py#L90-L91
        self.finish(json.dumps({
            "server_root": getattr(self.contents_manager, 'root_dir', None)
        }))

def setup_handlers(web_app):
    """
    Setups all of the git command handlers.
    Every handler is defined here, to be used in git.py file.
    """

    git_handlers = [
        ("/git/show_top_level", GitShowTopLevelHandler),
        ("/git/show_prefix", GitShowPrefixHandler),
        ("/git/add", GitAddHandler),
        ("/git/add_all_unstaged", GitAddAllUnstagedHandler),
        ("/git/add_all_untracked", GitAddAllUntrackedHandler),
        ("/git/status", GitStatusHandler),
        ("/git/branch", GitBranchHandler),
        ("/git/reset", GitResetHandler),
        ("/git/delete_commit", GitDeleteCommitHandler),
        ("/git/reset_to_commit", GitResetToCommitHandler),
        ("/git/checkout", GitCheckoutHandler),
        ("/git/commit", GitCommitHandler),
        ("/git/pull", GitPullHandler),
        ("/git/push", GitPushHandler),
        ("/git/diff", GitDiffHandler),
        ("/git/log", GitLogHandler),
        ("/git/detailed_log", GitDetailedLogHandler),
        ("/git/init", GitInitHandler),
        ("/git/all_history", GitAllHistoryHandler),
        ("/git/clone", GitCloneHandler),
        ("/git/upstream", GitUpstreamHandler),
        ("/git/config", GitConfigHandler),
        ("/git/changed_files", GitChangedFilesHandler),
        ("/git/server_root", GitServerRootHandler)
    ]

    # add the baseurl to our paths
    base_url = web_app.settings["base_url"]
    git_handlers = [(ujoin(base_url, x[0]), x[1]) for x in git_handlers]

    web_app.add_handlers(".*", git_handlers)



"""
Module for executing git commands, sending results back to the handlers
"""
import os
import subprocess
from subprocess import Popen, PIPE, CalledProcessError

import pexpect
from urllib.parse import unquote
from tornado.web import HTTPError


ALLOWED_OPTIONS = ['user.name', 'user.email']


class GitAuthInputWrapper:
    """
    Helper class which is meant to replace subprocess.Popen for communicating
    with git CLI when also sending username and password for auth
    """
    def __init__(self, command, cwd, env, username, password):
        self.command = command
        self.cwd = cwd
        self.env = env
        self.username = username
        self.password = password
    def communicate(self):
        try:
            p = pexpect.spawn(
                self.command, 
                cwd = self.cwd,
                env = self.env
            )

            # We expect a prompt from git
            # In most of cases git will prompt for username and
            #  then for password
            # In some cases (Bitbucket) username is included in
            #  remote URL, so git will not ask for username 
            i = p.expect(['Username for .*: ', 'Password for .*:'])
            if i==0: #ask for username then password
                p.sendline(self.username)
                p.expect('Password for .*:')
                p.sendline(self.password)
            elif i==1: #only ask for password
                p.sendline(self.password)

            p.expect(pexpect.EOF)
            response = p.before

            self.returncode = p.wait()
            p.close()
            
            return response
        except pexpect.exceptions.EOF: #In case of pexpect failure
            response = p.before
            self.returncode = p.exitstatus
            p.close() #close process
            return response


class Git:
    """
    A single parent class containing all of the individual git methods in it.
    """

    def __init__(self, root_dir, *args, **kwargs):
        super(Git, self).__init__(*args, **kwargs)
        self.root_dir = os.path.realpath(os.path.expanduser(root_dir))

    def config(self, top_repo_path, **kwargs):
        """Get or set Git options.
        
        If no kwargs, all options are returned. Otherwise kwargs are set.
        """
        response = {"code": 1}

        if len(kwargs):
            output = []
            for k, v in filter(lambda t: True if t[0] in ALLOWED_OPTIONS else False, kwargs.items()):
                cmd = ["git", "config", "--add", k, v]
                p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=top_repo_path)
                out, err = p.communicate()
                output.append(out.decode("utf-8").strip())
                response["code"] = p.returncode
                if p.returncode != 0:
                    response["command"] = " ".join(cmd)
                    response["message"] = err.decode("utf-8").strip()
                    return response

            response["message"] = "\n".join(output).strip()
        else:
            cmd = ["git", "config", "--list"]
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=top_repo_path)
            output, error = p.communicate()

            response = {"code": p.returncode}

            if p.returncode != 0:
                response["command"] = " ".join(cmd)
                response["message"] = error.decode("utf-8").strip()
            else:
                raw = output.decode("utf-8").strip()
                response["options"] = dict()
                for l in raw.split("\n"):
                    k, v = l.split("=", maxsplit=1)
                    if k in ALLOWED_OPTIONS:
                        response["options"][k] = v

        return response

    def changed_files(self, base=None, remote=None, single_commit=None):
        """Gets the list of changed files between two Git refs, or the files changed in a single commit

        There are two reserved "refs" for the base
            1. WORKING : Represents the Git working tree
            2. INDEX: Represents the Git staging area / index
        
        Keyword Arguments:
            single_commit {string} -- The single commit ref
            base {string} -- the base Git ref
            remote {string} -- the remote Git ref
        
        Returns:
            dict -- the response of format {
                "code": int, # Command status code
                "files": [string, string], # List of files changed.
                "message": [string] # Error response
            }
        """
        if single_commit:
            cmd = ['git', 'diff', '{}^!'.format(single_commit), '--name-only']
        elif base and remote:
            if base == 'WORKING':
                cmd = ['git', 'diff', remote, '--name-only']
            elif base == 'INDEX':
                cmd = ['git', 'diff', '--staged', remote, '--name-only']
            else:
                cmd = ['git', 'diff', base, remote, '--name-only']
        else:
            raise HTTPError(400, 'Either single_commit or (base and remote) must be provided')

        
        response = {}
        try:
            stdout = subprocess.check_output(
                cmd, 
                cwd=self.root_dir,
                stderr=subprocess.STDOUT
            )
            response['files'] = stdout.decode('utf-8').strip().split('\n')
            response['code'] = 0
        except CalledProcessError as e:
            response['message'] =  e.output.decode('utf-8')
            response['code'] = e.returncode

        return response


    def clone(self, current_path, repo_url, auth=None):
        """
        Execute `git clone`.
        When no auth is provided, disables prompts for the password to avoid the terminal hanging.
        When auth is provided, await prompts for username/passwords and sends them
        :param current_path: the directory where the clone will be performed.
        :param repo_url: the URL of the repository to be cloned.
        :param auth: OPTIONAL dictionary with 'username' and 'password' fields
        :return: response with status code and error message.
        """
        env = os.environ.copy()
        if (auth):
            env["GIT_TERMINAL_PROMPT"] = "1"
            p = GitAuthInputWrapper(
                command='git clone {} -q'.format(unquote(repo_url)),
                cwd=os.path.join(self.root_dir, current_path),
                env = env,
                username=auth['username'],
                password=auth['password'],
            )
            error = p.communicate()
        else:
            env["GIT_TERMINAL_PROMPT"] = "0"
            p = subprocess.Popen(
                ['git', 'clone', unquote(repo_url)],
                stdout=PIPE,
                stderr=PIPE,
                env = env,
                cwd=os.path.join(self.root_dir, current_path),
            )
            _, error = p.communicate()

        response = {"code": p.returncode}

        if p.returncode != 0:
            response["message"] = error.decode("utf-8").strip()

        return response

    def status(self, current_path):
        """
        Execute git status command & return the result.
        """
        p = Popen(
            ["git", "status", "--porcelain"],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = []
            line_array = my_output.decode("utf-8").splitlines()
            for line in line_array:
                to1 = None
                from_path = line[3:]
                if line[0] == "R":
                    to0 = line[3:].split(" -> ")
                    to1 = to0[len(to0) - 1]
                else:
                    to1 = line[3:]
                if to1.startswith('"'):
                    to1 = to1[1:]
                if to1.endswith('"'):
                    to1 = to1[:-1]
                result.append(
                    {"x": line[0], "y": line[1], "to": to1, "from": from_path}
                )
            return {"code": p.returncode, "files": result}
        else:
            return {
                "code": p.returncode,
                "command": "git status --porcelain",
                "message": my_error.decode("utf-8"),
            }

    def log(self, current_path, history_count=10):
        """
        Execute git log command & return the result.
        """
        p = Popen(
            ["git", "log", "--pretty=format:%H%n%an%n%ar%n%s", ("-%d" % history_count)],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = []
            line_array = my_output.decode("utf-8").splitlines()
            i = 0
            PREVIOUS_COMMIT_OFFSET = 4
            while i < len(line_array):
                if i + PREVIOUS_COMMIT_OFFSET < len(line_array):
                    result.append(
                        {
                            "commit": line_array[i],
                            "author": line_array[i + 1],
                            "date": line_array[i + 2],
                            "commit_msg": line_array[i + 3],
                            "pre_commit": line_array[i + PREVIOUS_COMMIT_OFFSET],
                        }
                    )
                else:
                    result.append(
                        {
                            "commit": line_array[i],
                            "author": line_array[i + 1],
                            "date": line_array[i + 2],
                            "commit_msg": line_array[i + 3],
                            "pre_commit": "",
                        }
                    )
                i += PREVIOUS_COMMIT_OFFSET
            return {"code": p.returncode, "commits": result}
        else:
            return {"code": p.returncode, "message": my_error.decode("utf-8")}

    def detailed_log(self, selected_hash, current_path):
        """
        Execute git log -1 --stat --numstat --oneline command (used to get
        insertions & deletions per file) & return the result.
        """
        p = subprocess.Popen(
            ["git", "log", "-1", "--stat", "--numstat", "--oneline", selected_hash],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = []
            note = [0] * 3
            count = 0
            temp = ""
            line_array = my_output.decode("utf-8").splitlines()
            length = len(line_array)
            INSERTION_INDEX = 0
            DELETION_INDEX = 1
            MODIFIED_FILE_PATH_INDEX = 2
            if length > 1:
                temp = line_array[length - 1]
                words = temp.split()
                for i in range(0, len(words)):
                    if words[i].isdigit():
                        note[count] = words[i]
                        count += 1
                for num in range(1, int(length / 2)):
                    line_info = line_array[num].split(maxsplit=2)
                    words = line_info[2].split("/")
                    length = len(words)
                    result.append(
                        {
                            "modified_file_path": line_info[MODIFIED_FILE_PATH_INDEX],
                            "modified_file_name": words[length - 1],
                            "insertion": line_info[INSERTION_INDEX],
                            "deletion": line_info[DELETION_INDEX],
                        }
                    )

            if note[2] == 0 and length > 1:
                if "-" in temp:
                    exchange = note[1]
                    note[1] = note[2]
                    note[2] = exchange

            return {
                "code": p.returncode,
                "modified_file_note": temp,
                "modified_files_count": note[0],
                "number_of_insertions": note[1],
                "number_of_deletions": note[2],
                "modified_files": result,
            }
        else:
            return {
                "code": p.returncode,
                "command": "git log_1",
                "message": my_error.decode("utf-8"),
            }

    def diff(self, top_repo_path):
        """
        Execute git diff command & return the result.
        """
        p = Popen(
            ["git", "diff", "--numstat"], stdout=PIPE, stderr=PIPE, cwd=top_repo_path
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = []
            line_array = my_output.decode("utf-8").splitlines()
            for line in line_array:
                linesplit = line.split()
                result.append(
                    {
                        "insertions": linesplit[0],
                        "deletions": linesplit[1],
                        "filename": linesplit[2],
                    }
                )
            return {"code": p.returncode, "result": result}
        else:
            return {"code": p.returncode, "message": my_error.decode("utf-8")}

    def branch(self, current_path):
        """
        Execute 'git for-each-ref' command & return the result.
        """
        heads = self.branch_heads(current_path)
        if heads["code"] != 0:
            # error; bail
            return heads

        remotes = self.branch_remotes(current_path)
        if remotes["code"] != 0:
            # error; bail
            return remotes

        # all's good; concatenate results and return
        return {"code": 0, "branches": heads["branches"] + remotes["branches"], "current_branch": heads["current_branch"]}

    def branch_heads(self, current_path):
        """
        Execute 'git for-each-ref' command on refs/heads & return the result.
        """
        # Format reference: https://git-scm.com/docs/git-for-each-ref#_field_names
        formats = ['refname:short', 'objectname', 'upstream:short', 'HEAD']
        cmd = ["git", "for-each-ref", "--format=" + "%09".join("%({})".format(f) for f in formats), "refs/heads/"]
        p = subprocess.Popen(
            cmd,
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        output, error = p.communicate()
        if p.returncode == 0:
            current_branch = None
            results = []
            try:
                for name,commit_sha,upstream_name,is_current_branch in (line.split('\t') for line in output.decode("utf-8").splitlines()):
                    is_current_branch = bool(is_current_branch.strip())

                    branch = {
                        "is_current_branch": is_current_branch,
                        "is_remote_branch": False,
                        "name": name,
                        "upstream": upstream_name if upstream_name else None,
                        "top_commit": commit_sha,
                        "tag": None,
                    }
                    results.append(branch)
                    if is_current_branch:
                        current_branch = branch

                # Above can fail in certain cases, such as an empty repo with
                # no commits. In that case, just fall back to determining
                # current branch
                if not current_branch:
                    branch = {
                        "is_current_branch": True,
                        "is_remote_branch": False,
                        "name": self.get_current_branch(current_path),
                        "upstream": None,
                        "top_commit": None,
                        "tag": None,
                    }
                    results.append(branch)
                    current_branch = branch

                return {"code": p.returncode, "branches": results, "current_branch": current_branch}

            except Exception as downstream_error:
                return {
                    "code": -1,
                    "command": ' '.join(cmd),
                    "message": str(downstream_error),
                }
        else:
            return {
                "code": p.returncode,
                "command": ' '.join(cmd),
                "message": error.decode("utf-8"),
            }

    def branch_remotes(self, current_path):
        """
        Execute 'git for-each-ref' command on refs/heads & return the result.
        """
        # Format reference: https://git-scm.com/docs/git-for-each-ref#_field_names
        formats = ['refname:short', 'objectname']
        cmd = ["git", "for-each-ref", "--format=" + "%09".join("%({})".format(f) for f in formats), "refs/remotes/"]
        p = subprocess.Popen(
            cmd,
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        output, error = p.communicate()
        if p.returncode == 0:
            results = []
            try:
                for name,commit_sha in (line.split('\t') for line in output.decode("utf-8").splitlines()):
                    results.append({
                        "is_current_branch": False,
                        "is_remote_branch": True,
                        "name": name,
                        "upstream": None,
                        "top_commit": commit_sha,
                        "tag": None,
                    })
                return {"code": p.returncode, "branches": results}
            except Exception as downstream_error:
                return {
                    "code": -1,
                    "command": ' '.join(cmd),
                    "message": str(downstream_error),
                }
        else:
            return {
                "code": p.returncode,
                "command": ' '.join(cmd),
                "message": error.decode("utf-8"),
            }

    def show_top_level(self, current_path):
        """
        Execute git --show-toplevel command & return the result.
        """
        p = Popen(
            ["git", "rev-parse", "--show-toplevel"],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = {
                "code": p.returncode,
                "top_repo_path": my_output.decode("utf-8").strip("\n"),
            }
            return result
        else:
            return {
                "code": p.returncode,
                "command": "git rev-parse --show-toplevel",
                "message": my_error.decode("utf-8"),
            }

    def show_prefix(self, current_path):
        """
        Execute git --show-prefix command & return the result.
        """
        p = Popen(
            ["git", "rev-parse", "--show-prefix"],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            result = {
                "code": p.returncode,
                "under_repo_path": my_output.decode("utf-8").strip("\n"),
            }
            return result
        else:
            return {
                "code": p.returncode,
                "command": "git rev-parse --show-prefix",
                "message": my_error.decode("utf-8"),
            }

    def add(self, filename, top_repo_path):
        """
        Execute git add<filename> command & return the result.
        """
        if not isinstance(filename, str):
            # assume filename is a sequence of str
            cmd = ["git", "add"] + list(filename)
        else:
            cmd = ["git", "add", filename]

        return subprocess.check_output(cmd, cwd=top_repo_path)

    def add_all(self, top_repo_path):
        """
        Execute git add all command & return the result.
        """
        my_output = subprocess.check_output(["git", "add", "-A"], cwd=top_repo_path)
        return my_output

    def add_all_unstaged(self, top_repo_path):
        """
        Execute git add all unstaged command & return the result.
        """
        e = 'git add -u'
        my_output = subprocess.call(e, shell=True, cwd=top_repo_path)
        return {"result": my_output}

    def add_all_untracked(self, top_repo_path):
        """
        Execute git add all untracked command & return the result.
        """
        e = 'echo "a\n*\nq\n" | git add -i'
        my_output = subprocess.call(e, shell=True, cwd=top_repo_path)
        return {"result": my_output}

    def reset(self, filename, top_repo_path):
        """
        Execute git reset <filename> command & return the result.
        """
        my_output = subprocess.check_output(
            ["git", "reset", "--", filename], cwd=top_repo_path
        )
        return my_output

    def reset_all(self, top_repo_path):
        """
        Execute git reset command & return the result.
        """
        my_output = subprocess.check_output(["git", "reset"], cwd=top_repo_path)
        return my_output

    def delete_commit(self, commit_id, top_repo_path):
        """
        Delete a specified commit from the repository.
        """
        my_output = subprocess.check_output(
            ["git", "revert", "--no-commit", commit_id], cwd=top_repo_path
        )
        return my_output

    def reset_to_commit(self, commit_id, top_repo_path):
        """
        Reset the current branch to a specific past commit.
        """
        cmd = ["git", "reset", "--hard"]
        if commit_id:
            cmd.append(commit_id)

        my_output = subprocess.check_output(
            cmd, cwd=top_repo_path
        )
        return my_output

    def checkout_new_branch(self, branchname, current_path):
        """
        Execute git checkout <make-branch> command & return the result.
        """
        p = Popen(
            ["git", "checkout", "-b", branchname],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            return {"code": p.returncode, "message": my_output.decode("utf-8")}
        else:
            return {
                "code": p.returncode,
                "command": "git checkout " + "-b" + branchname,
                "message": my_error.decode("utf-8"),
            }

    def _get_branch_reference(self, branchname, current_path):
        """
        Execute git rev-parse --symbolic-full-name <branch-name> and return the result (or None).
        """
        p = subprocess.Popen(
            ["git", "rev-parse", "--symbolic-full-name", branchname],
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        my_output, my_error = p.communicate()
        if p.returncode == 0:
            return my_output.decode("utf-8").strip("\n")
        else:
            return None

    def checkout_branch(self, branchname, current_path):
        """
        Execute git checkout <branch-name> command & return the result.
        Use the --track parameter for a remote branch.
        """
        reference_name = self._get_branch_reference(branchname, current_path)
        if reference_name is None:
            is_remote_branch = False
        else:
            is_remote_branch = self._is_remote_branch(reference_name)

        if is_remote_branch:
            cmd = ["git", "checkout", "--track", branchname]
        else:
            cmd = ["git", "checkout", branchname]

        p = subprocess.Popen(
            cmd,
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )

        my_output, my_error = p.communicate()
        if p.returncode == 0:
            return { "code": 0, "message": my_output.decode("utf-8") }
        else:
            return { "code": p.returncode, "message": my_error.decode("utf-8"), "command":  " ".join(cmd) }

    def checkout(self, filename, top_repo_path):
        """
        Execute git checkout command for the filename & return the result.
        """
        my_output = subprocess.check_output(
            ["git", "checkout", "--", filename], cwd=top_repo_path
        )
        return my_output

    def checkout_all(self, top_repo_path):
        """
        Execute git checkout command & return the result.
        """
        my_output = subprocess.check_output(
            ["git", "checkout", "--", "."], cwd=top_repo_path
        )
        return my_output

    def commit(self, commit_msg, top_repo_path):
        """
        Execute git commit <filename> command & return the result.
        """
        my_output = subprocess.check_output(
            ["git", "commit", "-m", commit_msg], cwd=top_repo_path
        )
        return my_output
            
    def pull(self, curr_fb_path, auth=None):
        """
        Execute git pull --no-commit.  Disables prompts for the password to avoid the terminal hanging while waiting
        for auth.
        """
        env = os.environ.copy()
        if (auth):
            env["GIT_TERMINAL_PROMPT"] = "1"
            p = GitAuthInputWrapper(
                command = 'git pull --no-commit',
                cwd = os.path.join(self.root_dir, curr_fb_path),
                env = env,
                username = auth['username'],
                password = auth['password']
            )
            error = p.communicate()
        else:
            env["GIT_TERMINAL_PROMPT"] = "0"
            p = subprocess.Popen(
                ['git', 'pull', '--no-commit'],
                stdout=PIPE,
                stderr=PIPE,
                env = env,
                cwd=os.path.join(self.root_dir, curr_fb_path),
            )
            _, error = p.communicate()

        response = {"code": p.returncode}

        if p.returncode != 0:
            response["message"] = error.decode("utf-8").strip()

        return response

    def push(self, remote, branch, curr_fb_path, auth=None):
        """
        Execute `git push $UPSTREAM $BRANCH`. The choice of upstream and branch is up to the caller.
        """
        env = os.environ.copy()
        if (auth):
            env["GIT_TERMINAL_PROMPT"] = "1"
            p = GitAuthInputWrapper(
                command = 'git push {} {}'.format(remote, branch),
                cwd = os.path.join(self.root_dir, curr_fb_path),
                env = env,
                username = auth['username'],
                password = auth['password']
            )
            error = p.communicate()
        else:
            env["GIT_TERMINAL_PROMPT"] = "0"
            p = subprocess.Popen(
                ['git', 'push', remote, branch],
                stdout=PIPE,
                stderr=PIPE,
                env = env,
                cwd=os.path.join(self.root_dir, curr_fb_path),
            )
            _, error = p.communicate()

        response = {"code": p.returncode}

        if p.returncode != 0:
            response["message"] = error.decode("utf-8").strip()

        return response

    def init(self, current_path):
        """
        Execute git init command & return the result.
        """
        my_output = subprocess.check_output(
            ["git", "init"], cwd=os.path.join(self.root_dir, current_path)
        )
        return my_output

    def _is_remote_branch(self, branch_reference):
        """Check if given branch is remote branch by comparing with 'remotes/',
        TODO : Consider a better way to check remote branch
        """
        return branch_reference.startswith("refs/remotes/")

    def _get_branch_name(self, branch_reference):
        """Get branch name for given branch
        """
        if branch_reference.startswith("refs/heads/"):
            return branch_reference.split("refs/heads/")[1]
        if branch_reference.startswith("refs/remotes/"):
            return branch_reference.split("refs/remotes/")[1]

        raise ValueError("Reference [{}] is not a valid branch.", branch_reference)

    def get_current_branch(self, current_path):
        """Use `symbolic-ref` to get the current branch name. In case of
        failure, assume that the HEAD is currently detached, and fall back
        to the `branch` command to get the name.
        See https://git-blame.blogspot.com/2013/06/checking-current-branch-programatically.html
        """
        command = ["git", "symbolic-ref", "HEAD"]
        p = subprocess.Popen(
            command,
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        output, error = p.communicate()
        if p.returncode == 0:
            return output.decode("utf-8").split('/')[-1].strip()
        elif "not a symbolic ref" in error.decode("utf-8").lower():
            return self._get_current_branch_detached(current_path)
        else:
            raise Exception(
                "Error [{}] occurred while executing [{}] command to get current branch.".format(
                    error.decode("utf-8"), " ".join(command)
                )
            )

    def _get_current_branch_detached(self, current_path):
        """Execute 'git branch -a' to get current branch details in case of detached HEAD
        """
        command = ["git", "branch", "-a"]
        p = subprocess.Popen(
            command,
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        output, error = p.communicate()
        if p.returncode == 0:
            for branch in output.decode("utf-8").splitlines():
                branch = branch.strip()
                if branch.startswith("*"):
                    return branch.lstrip("* ")
        else:
            raise Exception(
                "Error [{}] occurred while executing [{}] command to get detached HEAD name.".format(
                    error.decode("utf-8"), " ".join(command)
                )
            )

    def get_upstream_branch(self, current_path, branch_name):
        """Execute 'git rev-parse --abbrev-ref branch_name@{upstream}' to get
        upstream branch name tracked by given local branch.
        Reference : https://git-scm.com/docs/git-rev-parse#git-rev-parse-emltbranchnamegtupstreamemegemmasterupstreamememuem
        """
        command = [
            "git",
            "rev-parse",
            "--abbrev-ref",
            "{}@{{upstream}}".format(branch_name),
        ]
        p = subprocess.Popen(
            command,
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        output, error = p.communicate()
        if p.returncode == 0:
            return output.decode("utf-8").strip()
        elif "fatal: no upstream configured for branch" in error.decode("utf-8").lower():
            return None
        elif "unknown revision or path not in the working tree" in error.decode("utf-8").lower():
            return None
        else:
            raise Exception(
                "Error [{}] occurred while executing [{}] command to get upstream branch.".format(
                    error.decode("utf-8"), " ".join(command)
                )
            )

    def _get_tag(self, current_path, commit_sha):
        """Execute 'git describe commit_sha' to get
        nearest tag associated with lastest commit in branch.
        Reference : https://git-scm.com/docs/git-describe#git-describe-ltcommit-ishgt82308203
        """
        command = ["git", "describe", "--tags", commit_sha]
        p = subprocess.Popen(
            command,
            stdout=PIPE,
            stderr=PIPE,
            cwd=os.path.join(self.root_dir, current_path),
        )
        output, error = p.communicate()
        if p.returncode == 0:
            return output.decode("utf-8").strip()
        elif "fatal: no tags can describe '{}'.".format(commit_sha) in error.decode(
            "utf-8"
        ).lower():
            return None
        elif "fatal: no names found" in error.decode("utf-8").lower():
            return None
        else:
            raise Exception(
                "Error [{}] occurred while executing [{}] command to get nearest tag associated with branch.".format(
                    error.decode("utf-8"), " ".join(command)
                )
            )



from subprocess import PIPE

from mock import patch, call, Mock

from jupyterlab_git.git import Git, GitAuthInputWrapper


@patch('subprocess.Popen')
@patch('os.environ', {'TEST': 'test'})
def test_git_clone_success(mock_subproc_popen):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': ('output', 'error'.encode('utf-8')),
        'returncode': 0
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').clone(current_path='test_curr_path', repo_url='ghjkhjkl')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'clone', 'ghjkhjkl'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '0'},
        ),
        call().communicate()
    ])
    assert {'code': 0} == actual_response


@patch('subprocess.Popen')
@patch('os.environ', {'TEST': 'test'})
def test_git_clone_failure_from_git(mock_subproc_popen):
    """
    Git internally will throw an error if it is an invalid URL, or if there is a permissions issue. We want to just
    relay it back to the user.

    """
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': ('test_output', 'fatal: Not a git repository'.encode('utf-8')),
        'returncode': 128
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').clone(current_path='test_curr_path', repo_url='ghjkhjkl')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'clone', 'ghjkhjkl'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '0'},
        ),
        call().communicate()
    ])
    assert {'code': 128, 'message': 'fatal: Not a git repository'} == actual_response

@patch('jupyterlab_git.git.GitAuthInputWrapper')
@patch('os.environ', {'TEST': 'test'})
def test_git_clone_with_auth_success(mock_GitAuthInputWrapper):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': '',
        'returncode': 0
    }
    process_mock.configure_mock(**attrs)
    mock_GitAuthInputWrapper.return_value = process_mock

    # When
    auth = {
        'username' : 'asdf', 
        'password' : 'qwerty'
    }
    actual_response = Git(root_dir='/bin').clone(current_path='test_curr_path', repo_url='ghjkhjkl', auth=auth)

    # Then
    mock_GitAuthInputWrapper.assert_has_calls([
        call(
            command = 'git clone ghjkhjkl -q',
            cwd = '/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '1'},
            username = 'asdf',
            password = 'qwerty'
        ),
        call().communicate()
    ])
    assert {'code': 0} == actual_response

@patch('jupyterlab_git.git.GitAuthInputWrapper')
@patch('os.environ', {'TEST': 'test'})
def test_git_clone_with_auth_wrong_repo_url_failure_from_git(mock_GitAuthInputWrapper):
    """
    Git internally will throw an error if it is an invalid URL, or if there is a permissions issue. We want to just
    relay it back to the user.

    """
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': "fatal: repository 'ghjkhjkl' does not exist".encode('utf-8'),
        'returncode': 128
    }
    process_mock.configure_mock(**attrs)
    mock_GitAuthInputWrapper.return_value = process_mock

    # When
    auth = {
        'username' : 'asdf', 
        'password' : 'qwerty'
    }
    actual_response = Git(root_dir='/bin').clone(current_path='test_curr_path', repo_url='ghjkhjkl', auth=auth)

    # Then
    mock_GitAuthInputWrapper.assert_has_calls([
        call(
            command = 'git clone ghjkhjkl -q',
            cwd = '/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '1'},
            username = 'asdf',
            password = 'qwerty'
        ),
        call().communicate()
    ])
    assert {'code': 128, 'message': "fatal: repository 'ghjkhjkl' does not exist"} == actual_response

@patch('jupyterlab_git.git.GitAuthInputWrapper')
@patch('os.environ', {'TEST': 'test'})
def test_git_clone_with_auth_auth_failure_from_git(mock_GitAuthInputWrapper):
    """
    Git internally will throw an error if it is an invalid URL, or if there is a permissions issue. We want to just
    relay it back to the user.

    """
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': "remote: Invalid username or password.\r\nfatal: Authentication failed for 'ghjkhjkl'".encode('utf-8'),
        'returncode': 128
    }
    process_mock.configure_mock(**attrs)
    mock_GitAuthInputWrapper.return_value = process_mock

    # When
    auth = {
        'username' : 'asdf', 
        'password' : 'qwerty'
    }
    actual_response = Git(root_dir='/bin').clone(current_path='test_curr_path', repo_url='ghjkhjkl', auth=auth)

    # Then
    mock_GitAuthInputWrapper.assert_has_calls([
        call(
            command = 'git clone ghjkhjkl -q',
            cwd = '/bin/test_curr_path',
            env={'TEST': 'test', 'GIT_TERMINAL_PROMPT': '1'},
            username = 'asdf',
            password = 'qwerty'
        ),
        call().communicate()
    ])
    assert {'code': 128, 'message': "remote: Invalid username or password.\r\nfatal: Authentication failed for 'ghjkhjkl'"} == actual_response



import json
from mock import Mock, ANY, patch

from jupyterlab_git.handlers import (
    GitAllHistoryHandler,
    GitBranchHandler,
    GitLogHandler,
    GitPushHandler,
    GitUpstreamHandler,
    setup_handlers
)

def test_mapping_added():
    mock_web_app = Mock()
    mock_web_app.settings = {
        'base_url': 'nb_base_url'
    }
    setup_handlers(mock_web_app)

    mock_web_app.add_handlers.assert_called_once_with(".*", ANY)


@patch('jupyterlab_git.handlers.GitAllHistoryHandler.__init__', Mock(return_value=None))
@patch('jupyterlab_git.handlers.GitAllHistoryHandler.get_json_body', Mock(return_value={'current_path': 'test_path', 'history_count': 25}))
@patch('jupyterlab_git.handlers.GitAllHistoryHandler.git')
@patch('jupyterlab_git.handlers.GitAllHistoryHandler.finish')
def test_all_history_handler_localbranch(mock_finish, mock_git):
    # Given
    show_top_level = {'code': 0, 'foo': 'top_level'}
    branch = 'branch_foo'
    log = 'log_foo'
    status = 'status_foo'

    mock_git.show_top_level.return_value = show_top_level
    mock_git.branch.return_value = branch
    mock_git.log.return_value = log
    mock_git.status.return_value = status

    # When
    GitAllHistoryHandler().post()

    # Then
    mock_git.show_top_level.assert_called_with('test_path')
    mock_git.branch.assert_called_with('test_path')
    mock_git.log.assert_called_with('test_path', 25)
    mock_git.status.assert_called_with('test_path')

    mock_finish.assert_called_with(json.dumps({
        'code': show_top_level['code'],
        'data': {
            'show_top_level': show_top_level,
            'branch': branch,
            'log': log,
            'status': status,
        }
    }))


@patch('jupyterlab_git.handlers.GitBranchHandler.__init__', Mock(return_value=None))
@patch('jupyterlab_git.handlers.GitBranchHandler.get_json_body', Mock(return_value={'current_path': 'test_path'}))
@patch('jupyterlab_git.handlers.GitBranchHandler.git')
@patch('jupyterlab_git.handlers.GitBranchHandler.finish')
def test_branch_handler_localbranch(mock_finish, mock_git):
    # Given
    branch = {
        'code': 0,
        'branches': [
            {
                'is_current_branch': True,
                'is_remote_branch': False,
                'name': 'feature-foo',
                'upstream': 'origin/feature-foo',
                'top_commit': 'abcdefghijklmnopqrstuvwxyz01234567890123',
                'tag': None,
            },
            {
                'is_current_branch': False,
                'is_remote_branch': False,
                'name': 'master',
                'upstream': 'origin/master',
                'top_commit': 'abcdefghijklmnopqrstuvwxyz01234567890123',
                'tag': None,
            },
            {
                'is_current_branch': False,
                'is_remote_branch': False,
                'name': 'feature-bar',
                'upstream': None,
                'top_commit': '01234567899999abcdefghijklmnopqrstuvwxyz',
                'tag': None
            },
            {
                'is_current_branch': False,
                'is_remote_branch': True,
                'name': 'origin/feature-foo',
                'upstream': None,
                'top_commit': 'abcdefghijklmnopqrstuvwxyz01234567890123',
                'tag': None,
            },
            {
                'is_current_branch': False,
                'is_remote_branch': True,
                'name': 'origin/master',
                'upstream': None,
                'top_commit': 'abcdefghijklmnopqrstuvwxyz01234567890123',
                'tag': None,
            }
        ]
    }

    mock_git.branch.return_value = branch

    # When
    GitBranchHandler().post()

    # Then
    mock_git.branch.assert_called_with('test_path')

    mock_finish.assert_called_with(json.dumps({
        'code': 0,
        'branches': branch['branches']
    }))

@patch('jupyterlab_git.handlers.GitLogHandler.__init__', Mock(return_value=None))
@patch('jupyterlab_git.handlers.GitLogHandler.get_json_body', Mock(return_value={'current_path': 'test_path', 'history_count': 20}))
@patch('jupyterlab_git.handlers.GitLogHandler.git')
@patch('jupyterlab_git.handlers.GitLogHandler.finish')
def test_log_handler(mock_finish, mock_git):
    # Given
    log = {'code': 0, 'commits': []}
    mock_git.log.return_value = log

    # When
    GitLogHandler().post()

    # Then
    mock_git.log.assert_called_with('test_path', 20)
    mock_finish.assert_called_with(json.dumps(log))


@patch('jupyterlab_git.handlers.GitLogHandler.__init__', Mock(return_value=None))
@patch('jupyterlab_git.handlers.GitLogHandler.get_json_body', Mock(return_value={'current_path': 'test_path'}))
@patch('jupyterlab_git.handlers.GitLogHandler.git')
@patch('jupyterlab_git.handlers.GitLogHandler.finish')
def test_log_handler_no_history_count(mock_finish, mock_git):
    # Given
    log = {'code': 0, 'commits': []}
    mock_git.log.return_value = log

    # When
    GitLogHandler().post()

    # Then
    mock_git.log.assert_called_with('test_path', 25)
    mock_finish.assert_called_with(json.dumps(log))


@patch('jupyterlab_git.handlers.GitPushHandler.__init__', Mock(return_value=None))
@patch('jupyterlab_git.handlers.GitPushHandler.get_json_body', Mock(return_value={'current_path': 'test_path'}))
@patch('jupyterlab_git.handlers.GitPushHandler.git')
@patch('jupyterlab_git.handlers.GitPushHandler.finish')
def test_push_handler_localbranch(mock_finish, mock_git):
    # Given
    mock_git.get_current_branch.return_value = 'foo'
    mock_git.get_upstream_branch.return_value = 'localbranch'
    mock_git.push.return_value = {'code': 0}

    # When
    GitPushHandler().post()

    # Then
    mock_git.get_current_branch.assert_called_with('test_path')
    mock_git.get_upstream_branch.assert_called_with('test_path', 'foo')
    mock_git.push.assert_called_with('.', 'HEAD:localbranch', 'test_path', None)
    mock_finish.assert_called_with('{"code": 0}')


@patch('jupyterlab_git.handlers.GitPushHandler.__init__', Mock(return_value=None))
@patch('jupyterlab_git.handlers.GitPushHandler.get_json_body', Mock(return_value={'current_path': 'test_path'}))
@patch('jupyterlab_git.handlers.GitPushHandler.git')
@patch('jupyterlab_git.handlers.GitPushHandler.finish')
def test_push_handler_remotebranch(mock_finish, mock_git):
    # Given
    mock_git.get_current_branch.return_value = 'foo'
    mock_git.get_upstream_branch.return_value = 'origin/remotebranch'
    mock_git.push.return_value = {'code': 0}

    # When
    GitPushHandler().post()

    # Then
    mock_git.get_current_branch.assert_called_with('test_path')
    mock_git.get_upstream_branch.assert_called_with('test_path', 'foo')
    mock_git.push.assert_called_with('origin', 'HEAD:remotebranch', 'test_path', None)
    mock_finish.assert_called_with('{"code": 0}')


@patch('jupyterlab_git.handlers.GitPushHandler.__init__', Mock(return_value=None))
@patch('jupyterlab_git.handlers.GitPushHandler.get_json_body', Mock(return_value={'current_path': 'test_path'}))
@patch('jupyterlab_git.handlers.GitPushHandler.git')
@patch('jupyterlab_git.handlers.GitPushHandler.finish')
def test_push_handler_noupstream(mock_finish, mock_git):
    # Given
    mock_git.get_current_branch.return_value = 'foo'
    mock_git.get_upstream_branch.return_value = ''
    mock_git.push.return_value = {'code': 0}

    # When
    GitPushHandler().post()

    # Then
    mock_git.get_current_branch.assert_called_with('test_path')
    mock_git.get_upstream_branch.assert_called_with('test_path', 'foo')
    mock_git.push.assert_not_called()
    mock_finish.assert_called_with('{"code": 128, "message": "fatal: The current branch foo has no upstream branch."}')


@patch('jupyterlab_git.handlers.GitUpstreamHandler.__init__', Mock(return_value=None))
@patch('jupyterlab_git.handlers.GitUpstreamHandler.get_json_body', Mock(return_value={'current_path': 'test_path'}))
@patch('jupyterlab_git.handlers.GitUpstreamHandler.git')
@patch('jupyterlab_git.handlers.GitUpstreamHandler.finish')
def test_upstream_handler_localbranch(mock_finish, mock_git):
    # Given
    mock_git.get_current_branch.return_value = 'foo'
    mock_git.get_upstream_branch.return_value = 'bar'

    # When
    GitUpstreamHandler().post()

    # Then
    mock_git.get_current_branch.assert_called_with('test_path')
    mock_git.get_upstream_branch.assert_called_with('test_path', 'foo')
    mock_finish.assert_called_with('{"upstream": "bar"}')



"""
Initialize the backend server extension
"""
from jupyterlab_git.handlers import setup_handlers
from jupyterlab_git.git import Git


def _jupyter_server_extension_paths():
    """
    Declare the Jupyter server extension paths.
    """
    return [{"module": "jupyterlab_git"}]


def _jupyter_nbextension_paths():
    """
    Declare the Jupyter notebook extension paths.
    """
    return [{"section": "notebook", "dest": "jupyterlab_git"}]


def load_jupyter_server_extension(nbapp):
    """
    Load the Jupyter server extension.
    """
    git = Git(nbapp.web_app.settings.get('server_root_dir'))
    nbapp.web_app.settings["git"] = git
    setup_handlers(nbapp.web_app)



# python lib
import pytest
from subprocess import PIPE
from mock import patch, call, Mock

# local lib
from jupyterlab_git.git import Git


def test_is_remote_branch():
    test_cases = [
        ('refs/heads/feature-foo', False),
        ('refs/heads/master', False),
        ('refs/remotes/origin/feature-foo', True),
        ('refs/remotes/origin/HEAD', True),
        ('refs/stash', False),
        ('refs/tags/v0.1.0', False),
        ('refs/tags/blah@0.2.0', False)
    ]
    for test_case in test_cases:
        actual_response = Git(root_dir='/bin')._is_remote_branch(test_case[0])
        assert test_case[1] == actual_response


def test_get_branch_name():
    good_test_cases = [
        ('refs/heads/feature-foo', 'feature-foo'),
        ('refs/heads/master', 'master'),
        ('refs/remotes/origin/feature-foo', 'origin/feature-foo'),
        ('refs/remotes/origin/HEAD', 'origin/HEAD')
    ]
    bad_test_cases = [
        'refs/stash',
        'refs/tags/v0.1.0',
        'refs/tags/blah@0.2.0'
    ]
    for test_case in good_test_cases:
        actual_response = Git(root_dir='/bin')._get_branch_name(test_case[0])
        assert test_case[1] == actual_response

    for test_case in bad_test_cases:
        with pytest.raises(ValueError):
            Git(root_dir='/bin')._get_branch_name(test_case)


@patch('subprocess.Popen')
def test_get_current_branch_success(mock_subproc_popen):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': ('feature-foo'.encode('utf-8'), ''.encode('utf-8')),
        'returncode': 0
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').get_current_branch(
        current_path='test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'symbolic-ref', 'HEAD'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ])
    assert 'feature-foo' == actual_response

@patch('subprocess.Popen')
@patch.object(Git, '_get_branch_reference', return_value=None)
def test_checkout_branch_noref_success(mock__get_branch_reference, mock_subproc_popen):
    branch='test-branch'
    curr_path='test_curr_path'
    stdout_message='checkout output from git'
    stderr_message=''
    rc=0

    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': (stdout_message.encode('utf-8'), stderr_message.encode('utf-8')),
        'returncode': rc
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').checkout_branch(branchname=branch, current_path=curr_path)

    # Then
    mock__get_branch_reference.assert_has_calls([ call(branch, curr_path) ])

    cmd=['git', 'checkout', branch]
    mock_subproc_popen.assert_has_calls([
        call(cmd, stdout=PIPE, stderr=PIPE, cwd='/bin/{}'.format(curr_path)),
        call().communicate()
    ])

    assert { "code": rc, "message": stdout_message } == actual_response


@patch('subprocess.Popen')
@patch.object(Git, '_get_branch_reference', return_value=None)
def test_checkout_branch_noref_failure(mock__get_branch_reference, mock_subproc_popen):
    branch='test-branch'
    curr_path='test_curr_path'
    stdout_message=''
    stderr_message="error: pathspec '{}' did not match any file(s) known to git".format(branch)
    rc=1

    # Given
    process_mock = Mock()
    attrs = { 'communicate.return_value': (stdout_message.encode('utf-8'), stderr_message.encode('utf-8')), 'returncode': rc }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').checkout_branch(branchname=branch, current_path=curr_path)

    # Then
    mock__get_branch_reference.assert_has_calls([ call(branch, curr_path) ])

    cmd=['git', 'checkout', branch]
    mock_subproc_popen.assert_has_calls([
        call(cmd, stdout=PIPE, stderr=PIPE, cwd='/bin/{}'.format(curr_path)),
        call().communicate()
    ])

    assert { "code": rc, "message": stderr_message,  "command": ' '.join(cmd) } == actual_response


@patch('subprocess.Popen')
@patch.object(Git, '_get_branch_reference', return_value="refs/remotes/remote_branch")
def test_checkout_branch_remoteref_success(mock__get_branch_reference, mock_subproc_popen):
    branch='test-branch'
    curr_path='test_curr_path'
    stdout_message='checkout output from git'
    stderr_message=''
    rc=0

    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': (stdout_message.encode('utf-8'), stderr_message.encode('utf-8')),
        'returncode': rc
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').checkout_branch(branchname=branch, current_path=curr_path)

    # Then
    mock__get_branch_reference.assert_has_calls([ call(branch, curr_path) ])

    cmd=['git', 'checkout', '--track', branch]
    mock_subproc_popen.assert_has_calls([
        call(cmd, stdout=PIPE, stderr=PIPE, cwd='/bin/{}'.format(curr_path)),
        call().communicate()
    ])
    assert { "code": rc, "message": stdout_message } == actual_response


@patch('subprocess.Popen')
@patch.object(Git, '_get_branch_reference', return_value="refs/heads/local_branch")
def test_checkout_branch_headsref_failure(mock__get_branch_reference, mock_subproc_popen):
    branch='test-branch'
    curr_path='test_curr_path'
    stdout_message=''
    stderr_message="error: pathspec '{}' did not match any file(s) known to git".format(branch)
    rc=1

    # Given
    process_mock = Mock()
    attrs = { 'communicate.return_value': (stdout_message.encode('utf-8'), stderr_message.encode('utf-8')), 'returncode': rc }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').checkout_branch(branchname=branch, current_path=curr_path)

    # Then
    mock__get_branch_reference.assert_has_calls([ call(branch, curr_path) ])

    cmd=['git', 'checkout', branch]
    mock_subproc_popen.assert_has_calls([
        call(cmd, stdout=PIPE, stderr=PIPE, cwd='/bin/{}'.format(curr_path)),
        call().communicate()
    ])
    assert { "code": rc, "message": stderr_message,  "command": ' '.join(cmd) } == actual_response


@patch('subprocess.Popen')
@patch.object(Git, '_get_branch_reference', return_value="refs/heads/local_branch")
def test_checkout_branch_headsref_success(mock__get_branch_reference, mock_subproc_popen):
    branch='test-branch'
    stdout_message='checkout output from git'
    stderr_message=''
    rc=0

    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': (stdout_message.encode('utf-8'), stderr_message.encode('utf-8')),
        'returncode': rc
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').checkout_branch(
        branchname=branch,
        current_path='test_curr_path')

    # Then
    cmd=['git', 'checkout', branch]
    mock_subproc_popen.assert_has_calls([
        call(cmd, stdout=PIPE, stderr=PIPE, cwd='/bin/test_curr_path'),
        call().communicate()
    ])
    assert { "code": rc, "message": stdout_message } == actual_response


@patch('subprocess.Popen')
@patch.object(Git, '_get_branch_reference', return_value="refs/remotes/remote_branch")
def test_checkout_branch_remoteref_failure(mock__get_branch_reference, mock_subproc_popen):
    branch='test-branch'
    stdout_message=''
    stderr_message="error: pathspec '{}' did not match any file(s) known to git".format(branch)
    rc=1

    # Given
    process_mock = Mock()
    attrs = { 'communicate.return_value': (stdout_message.encode('utf-8'), stderr_message.encode('utf-8')), 'returncode': rc }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin').checkout_branch(branchname=branch, current_path='test_curr_path')

    # Then
    cmd=['git', 'checkout', '--track', branch]
    mock_subproc_popen.assert_has_calls([
        call(cmd, stdout=PIPE, stderr=PIPE, cwd='/bin/test_curr_path'),
        call().communicate()
    ])
    assert { "code": rc, "message": stderr_message,  "command": ' '.join(cmd) } == actual_response



@patch('subprocess.Popen')
def test_get_branch_reference_success(mock_subproc_popen):
    actual_response = 0
    branch='test-branch'
    reference = 'refs/remotes/origin/test_branch'
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': (reference.encode('utf-8'), ''.encode('utf-8')),
        'returncode': 0
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock
 
    # When
    actual_response = Git(root_dir='/bin')._get_branch_reference(
        branchname=branch,
        current_path='test_curr_path')
 
    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'rev-parse', '--symbolic-full-name', branch],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ])
    assert actual_response == reference


@patch('subprocess.Popen')
def test_get_branch_reference_failure(mock_subproc_popen):
    actual_response = 0
    branch='test-branch'
    reference = 'test-branch'
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': (
            reference.encode('utf-8'),
            "fatal: ambiguous argument '{}': unknown revision or path not in the working tree.".format(branch).encode('utf-8')
        ),
        'returncode': 128
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock
 
    # When
    actual_response = Git(root_dir='/bin')._get_branch_reference(
        branchname=branch,
        current_path='test_curr_path')
 
    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'rev-parse', '--symbolic-full-name', branch],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ])
    assert actual_response is None


@patch('subprocess.Popen')
def test_get_current_branch_failure(mock_subproc_popen):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': (
            '', 'fatal: Not a git repository (or any of the parent directories): .git'.encode('utf-8')),
        'returncode': 128
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    with pytest.raises(Exception) as error:
        Git(root_dir='/bin').get_current_branch(current_path='test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'symbolic-ref', 'HEAD'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ])
    assert 'Error [fatal: Not a git repository (or any of the parent directories): .git] ' \
           'occurred while executing [git symbolic-ref HEAD] command to get current branch.' == str(
        error.value)


@patch('subprocess.Popen')
def test_get_current_branch_detached_success(mock_subproc_popen):
    # Given
    process_output = [
        '* (HEAD detached at origin/feature-foo)',
        '  master',
        '  remotes/origin/feature-foo',
        '  remotes/origin/HEAD'
    ]
    process_mock = Mock()
    attrs = {
        'communicate.return_value': ('\n'.join(process_output).encode('utf-8'), ''.encode('utf-8')),
        'returncode': 0
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin')._get_current_branch_detached(
        current_path='test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'branch', '-a'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ])
    assert '(HEAD detached at origin/feature-foo)' == actual_response


@patch('subprocess.Popen')
def test_get_current_branch_detached_failure(mock_subproc_popen):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': (
            '', 'fatal: Not a git repository (or any of the parent directories): .git'.encode('utf-8')),
        'returncode': 128
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    with pytest.raises(Exception) as error:
        Git(root_dir='/bin')._get_current_branch_detached(current_path='test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'branch', '-a'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ])
    assert 'Error [fatal: Not a git repository (or any of the parent directories): .git] ' \
           'occurred while executing [git branch -a] command to get detached HEAD name.' == str(error.value)


@patch('subprocess.Popen')
def test_get_upstream_branch_success(mock_subproc_popen):
    # Tuple structure : (branch_name, upstream_branch_name)
    test_cases = [
        ('feature-foo', 'origin/master'),
        ('master', 'origin/master'),
        ('feature-bar', 'feature-foo')
    ]

    for test_case in test_cases:
        # Given
        process_mock = Mock()
        attrs = {
            'communicate.return_value': (test_case[1].encode('utf-8'), ''.encode('utf-8')),
            'returncode': 0
        }
        process_mock.configure_mock(**attrs)
        mock_subproc_popen.return_value = process_mock

        # When
        actual_response = Git(root_dir='/bin').get_upstream_branch(
            current_path='test_curr_path', branch_name=test_case[0])

        # Then
        mock_subproc_popen.assert_has_calls([
            call(
                ['git', 'rev-parse', '--abbrev-ref',
                 '{}@{{upstream}}'.format(test_case[0])],
                stdout=PIPE,
                stderr=PIPE,
                cwd='/bin/test_curr_path'
            ),
            call().communicate()
        ])
        assert test_case[1] == actual_response
        mock_subproc_popen.reset_mock()
        process_mock.reset_mock()


@patch('subprocess.Popen')
def test_get_upstream_branch_failure(mock_subproc_popen):
    # Given
    process_mock = Mock(returncode=128)
    process_mock.communicate.side_effect = [
        ('', "fatal: no such branch: 'blah'".encode('utf-8')),
        ('', "fatal: no upstream configured for branch".encode('utf-8')),
        ('', "fatal: ambiguous argument 'blah@origin': unknown revision or path not in the working tree.".encode('utf-8'))
    ]
    mock_subproc_popen.return_value = process_mock

    # When: fatal: no such branch: 'blah'
    with pytest.raises(Exception) as error:
        Git(root_dir='/bin').get_upstream_branch(
            current_path='test_curr_path', branch_name='blah')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'rev-parse', '--abbrev-ref',
             'blah@{upstream}'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ], any_order=False)
    assert "Error [fatal: no such branch: 'blah'] " \
           "occurred while executing [git rev-parse --abbrev-ref blah@{upstream}] command to get upstream branch." == str(
        error.value)

    # When: fatal: no upstream configured for branch
    actual_response = Git(root_dir='/bin').get_upstream_branch(
        current_path='test_curr_path', branch_name='test')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'rev-parse', '--abbrev-ref',
             'test@{upstream}'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ], any_order=False)
    assert None == actual_response

    # When: "fatal: ambiguous argument 'blah@origin': unknown revision or path not in the working tree.
    actual_response = Git(root_dir='/bin').get_upstream_branch(
        current_path='test_curr_path', branch_name='blah')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'rev-parse', '--abbrev-ref',
             'blah@{upstream}'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ], any_order=False)
    assert None == actual_response


@patch('subprocess.Popen')
def test_get_tag_success(mock_subproc_popen):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': ('v0.3.0'.encode('utf-8'), ''.encode('utf-8')),
        'returncode': 0
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin')._get_tag(
        current_path='test_curr_path', commit_sha='abcdefghijklmnopqrstuvwxyz01234567890123')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'describe', '--tags', 'abcdefghijklmnopqrstuvwxyz01234567890123'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ])
    assert 'v0.3.0' == actual_response


@patch('subprocess.Popen')
def test_get_tag_failure(mock_subproc_popen):
    # Given
    process_mock = Mock(returncode=128)
    process_mock.communicate.side_effect = [
        ('', "fatal: Not a valid object name blah".encode('utf-8')),
        ('', "fatal: No tags can describe '01234567899999abcdefghijklmnopqrstuvwxyz'.".encode('utf-8'))
    ]
    mock_subproc_popen.return_value = process_mock

    # When
    with pytest.raises(Exception) as error:
        Git(root_dir='/bin')._get_tag(
            current_path='test_curr_path', commit_sha='blah')

    assert "Error [fatal: Not a valid object name blah] " \
           "occurred while executing [git describe --tags blah] command to get nearest tag associated with branch." == str(
        error.value)

    actual_response = Git(root_dir='/bin')._get_tag(
        current_path='test_curr_path', commit_sha='01234567899999abcdefghijklmnopqrstuvwxyz')

    assert None == actual_response

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'describe', '--tags', 'blah'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate(),
        call(
            ['git', 'describe', '--tags', '01234567899999abcdefghijklmnopqrstuvwxyz'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ], any_order=False)


@patch('subprocess.Popen')
def test_no_tags(mock_subproc_popen):
    # Given
    process_mock = Mock()
    attrs = {
        'communicate.return_value': (b'', b'fatal: No names found, cannot describe anything.\n'),
        'returncode': 128
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # When
    actual_response = Git(root_dir='/bin')._get_tag('/path/foo', '768c79ad661598889f29bdf8916f4cc488f5062a')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            ['git', 'describe', '--tags', '768c79ad661598889f29bdf8916f4cc488f5062a'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/path/foo'
        ),
        call().communicate()
    ])
    assert actual_response is None


@patch('subprocess.Popen')
def test_branch_success(mock_subproc_popen):
    # Given
    process_output_heads = [
        'feature-foo\tabcdefghijklmnopqrstuvwxyz01234567890123\torigin/feature-foo\t*',
        'master\tabcdefghijklmnopqrstuvwxyz01234567890123\torigin/master\t ',
        'feature-bar\t01234567899999abcdefghijklmnopqrstuvwxyz\t\t ',
    ]
    process_output_remotes = [
        'origin/feature-foo\tabcdefghijklmnopqrstuvwxyz01234567890123',
        'origin/master\tabcdefghijklmnopqrstuvwxyz01234567890123',
    ]
    process_mock = Mock(returncode=0)
    process_mock.communicate.side_effect = [
        # Response for get all refs/heads
        ('\n'.join(process_output_heads).encode('utf-8'), ''.encode('utf-8')),

        # Response for get all refs/remotes
        ('\n'.join(process_output_remotes).encode('utf-8'), ''.encode('utf-8')),
    ]
    mock_subproc_popen.return_value = process_mock

    expected_response = {
        'code': 0,
        'branches': [
            {
                'is_current_branch': True,
                'is_remote_branch': False,
                'name': 'feature-foo',
                'upstream': 'origin/feature-foo',
                'top_commit': 'abcdefghijklmnopqrstuvwxyz01234567890123',
                'tag': None,
            },
            {
                'is_current_branch': False,
                'is_remote_branch': False,
                'name': 'master',
                'upstream': 'origin/master',
                'top_commit': 'abcdefghijklmnopqrstuvwxyz01234567890123',
                'tag': None,
            },
            {
                'is_current_branch': False,
                'is_remote_branch': False,
                'name': 'feature-bar',
                'upstream': None,
                'top_commit': '01234567899999abcdefghijklmnopqrstuvwxyz',
                'tag': None
            },
            {
                'is_current_branch': False,
                'is_remote_branch': True,
                'name': 'origin/feature-foo',
                'upstream': None,
                'top_commit': 'abcdefghijklmnopqrstuvwxyz01234567890123',
                'tag': None,
            },
            {
                'is_current_branch': False,
                'is_remote_branch': True,
                'name': 'origin/master',
                'upstream': None,
                'top_commit': 'abcdefghijklmnopqrstuvwxyz01234567890123',
                'tag': None,
            }
        ],
        'current_branch': {
            'is_current_branch': True,
            'is_remote_branch': False,
            'name': 'feature-foo',
            'upstream': 'origin/feature-foo',
            'top_commit': 'abcdefghijklmnopqrstuvwxyz01234567890123',
            'tag': None,
        }
    }

    # When
    actual_response = Git(root_dir='/bin').branch(
        current_path='test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        # call to get refs/heads
        call(
            ["git", "for-each-ref", "--format=%(refname:short)%09%(objectname)%09%(upstream:short)%09%(HEAD)", "refs/heads/"],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate(),

        # call to get refs/remotes
        call(
            ["git", "for-each-ref", "--format=%(refname:short)%09%(objectname)", "refs/remotes/"],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ], any_order=False)

    assert expected_response == actual_response


@patch('subprocess.Popen')
def test_branch_failure(mock_subproc_popen):
    # Given
    expected_cmd = ["git", "for-each-ref", "--format=%(refname:short)%09%(objectname)%09%(upstream:short)%09%(HEAD)", "refs/heads/"]
    process_mock = Mock()
    attrs = {
        'communicate.return_value': (
            '', 'fatal: Not a git repository (or any of the parent directories): .git'.encode('utf-8')),
        'returncode': 128
    }
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock
    expected_response = {
        'code': 128,
        'command': ' '.join(expected_cmd),
        'message': 'fatal: Not a git repository (or any of the parent directories): .git',
    }

    # When
    actual_response = Git(root_dir='/bin').branch(current_path='test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        call(
            expected_cmd,
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ])

    assert expected_response == actual_response


@patch('subprocess.Popen')
def test_branch_success_detached_head(mock_subproc_popen):
    # Given
    process_output_heads = [
        'master\tabcdefghijklmnopqrstuvwxyz01234567890123\torigin/master\t '
    ]
    process_output_remotes = [
        'origin/feature-foo\tabcdefghijklmnopqrstuvwxyz01234567890123'
    ]
    detached_head_output = [
        '* (HEAD detached at origin/feature-foo)',
        '  master',
        '  remotes/origin/feature-foo'
    ]

    process_mock = Mock()
    com_returncodes = [0, 128, 0, 0]
    com_returns = [
        # Response for get all refs/heads
        ('\n'.join(process_output_heads).encode('utf-8'), ''.encode('utf-8')),
        # Response for get current branch
        (''.encode('utf-8'), 'fatal: ref HEAD is not a symbolic ref'.encode('utf-8')),
        # Response for get current branch detached
        ('\n'.join(detached_head_output).encode('utf-8'), ''.encode('utf-8')),
        # Response for get all refs/remotes
        ('\n'.join(process_output_remotes).encode('utf-8'), ''.encode('utf-8')),
    ]
    def com_mock_side_effect():
        process_mock.returncode = com_returncodes.pop(0)
        return com_returns.pop(0)
    process_mock.communicate.side_effect = com_mock_side_effect
    mock_subproc_popen.return_value = process_mock

    expected_response = {
        'code': 0,
        'branches': [
            {
                'is_current_branch': False,
                'is_remote_branch': False,
                'name': 'master',
                'upstream': 'origin/master',
                'top_commit': 'abcdefghijklmnopqrstuvwxyz01234567890123',
                'tag': None,
            },
            {
                'is_current_branch': True,
                'is_remote_branch': False,
                'name': '(HEAD detached at origin/feature-foo)',
                'upstream': None,
                'top_commit': None,
                'tag': None,
            },
            {
                'is_current_branch': False,
                'is_remote_branch': True,
                'name': 'origin/feature-foo',
                'upstream': None,
                'top_commit': 'abcdefghijklmnopqrstuvwxyz01234567890123',
                'tag': None,
            }
        ],
        'current_branch': {
            'is_current_branch': True,
            'is_remote_branch': False,
            'name': '(HEAD detached at origin/feature-foo)',
            'upstream': None,
            'top_commit': None,
            'tag': None,
        }
    }

    # When
    actual_response = Git(root_dir='/bin').branch(
        current_path='test_curr_path')

    # Then
    mock_subproc_popen.assert_has_calls([
        # call to get refs/heads
        call(
            ["git", "for-each-ref", "--format=%(refname:short)%09%(objectname)%09%(upstream:short)%09%(HEAD)", "refs/heads/"],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate(),

        # call to get current branch
        call(
            ['git', 'symbolic-ref', 'HEAD'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate(),

        # call to get current branch name given a detached head
        call(
            ['git', 'branch', '-a'],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate(),

        # call to get refs/remotes
        call(
            ["git", "for-each-ref", "--format=%(refname:short)%09%(objectname)", "refs/remotes/"],
            stdout=PIPE,
            stderr=PIPE,
            cwd='/bin/test_curr_path'
        ),
        call().communicate()
    ], any_order=False)

    assert expected_response == actual_response



"""Setup Module to setup Python serverextension for the jupyterlab git
extension. For non-dev installs, will also automatically
build (if package.json is present) and install (if the labextension exists,
eg the build succeeded) the corresponding labextension.
"""
from pathlib import Path
from subprocess import CalledProcessError

from setupbase import (
    command_for_func, create_cmdclass, ensure_python,
    get_version, HERE, run
)

import setuptools

# The name of the project
name='jupyterlab_git'

# Ensure a valid python version
ensure_python('>=3.5')

# Get our version
version = get_version(str(Path(name) / '_version.py'))

lab_path = Path(HERE) / name / 'labextension'

data_files_spec = [
    ('share/jupyter/lab/extensions', str(lab_path), '*.tgz'),
    ('etc/jupyter/jupyter_notebook_config.d',
     'jupyter-config/jupyter_notebook_config.d', 'jupyterlab_git.json'),
]

def runPackLabextension():
    if Path('package.json').is_file():
        try:
            run(['jlpm', 'build:labextension'])
        except CalledProcessError:
            pass
pack_labext = command_for_func(runPackLabextension)

cmdclass = create_cmdclass('pack_labext', data_files_spec=data_files_spec)
cmdclass['pack_labext'] = pack_labext
cmdclass.pop('develop')

with open("README.md", "r") as fh:
    long_description = fh.read()

setup_args = dict(
    name            = name,
    description     = "A server extension for JupyterLab's git extension",
    long_description= long_description,
    long_description_content_type="text/markdown",
    version         = version,
    cmdclass        = cmdclass,
    packages        = setuptools.find_packages(),
    author          = 'Jupyter Development Team',
    url             = 'https://github.com/jupyterlab/jupyterlab-git',
    license         = 'BSD',
    platforms       = "Linux, Mac OS X, Windows",
    keywords        = ['Jupyter', 'JupyterLab', 'Git'],
    classifiers     = [
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Framework :: Jupyter',
    ],
    install_requires = [
        'notebook',
        'nbdime >= 1.1.0',
        'pexpect'
    ],
    extras_require = {
        'test': [
            'pytest',
            'jupyterlab~=1.1',
        ],
    },
)

setuptools.setup(**setup_args)



import json
import subprocess

from mock import patch, call, Mock
import pytest

from jupyterlab_git.git import Git
from jupyterlab_git.handlers import GitConfigHandler


@patch("jupyterlab_git.handlers.GitConfigHandler.__init__", Mock(return_value=None))
@patch(
    "jupyterlab_git.handlers.GitConfigHandler.get_json_body",
    Mock(return_value={"path": "test_path"}),
)
@patch("jupyterlab_git.handlers.GitConfigHandler.git", Git("/bin"))
@patch("jupyterlab_git.handlers.GitConfigHandler.finish")
@patch("subprocess.Popen")
def test_git_get_config_success(popen, finish):
    # Given
    process_mock = Mock()
    attrs = {
        "communicate": Mock(
            return_value=(
                b"user.name=John Snow\nuser.email=john.snow@iscoming.com",
                b"",
            )
        ),
        "returncode": 0,
    }
    process_mock.configure_mock(**attrs)
    popen.return_value = process_mock

    # When
    GitConfigHandler().post()

    # Then
    popen.assert_called_once_with(
        ["git", "config", "--list"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd="test_path",
    )
    process_mock.communicate.assert_called_once_with()

    finish.assert_called_once_with(
        json.dumps(
            {
                "code": 0,
                "options": {
                    "user.name": "John Snow",
                    "user.email": "john.snow@iscoming.com",
                },
            }
        )
    )


@patch("jupyterlab_git.handlers.GitConfigHandler.__init__", Mock(return_value=None))
@patch(
    "jupyterlab_git.handlers.GitConfigHandler.get_json_body",
    Mock(
        return_value={
            "path": "test_path",
            "options": {
                "user.name": "John Snow",
                "user.email": "john.snow@iscoming.com",
            },
        }
    ),
)
@patch("jupyterlab_git.handlers.GitConfigHandler.git", Git("/bin"))
@patch("jupyterlab_git.handlers.GitConfigHandler.finish")
@patch("subprocess.Popen")
def test_git_set_config_success(popen, finish):
    # Given
    process_mock = Mock()
    attrs = {"communicate": Mock(return_value=(b"", b"")), "returncode": 0}
    process_mock.configure_mock(**attrs)
    popen.return_value = process_mock

    # When
    GitConfigHandler().post()

    # Then
    assert popen.call_count == 2
    assert (
        call(
            ["git", "config", "--add", "user.email", "john.snow@iscoming.com"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd="test_path",
        )
        in popen.call_args_list
    )
    assert (
        call(
            ["git", "config", "--add", "user.name", "John Snow"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd="test_path",
        )
        in popen.call_args_list
    )
    assert process_mock.communicate.call_count == 2

    finish.assert_called_once_with(json.dumps({"code": 0, "message": ""}))
